from dataclasses import dataclass
from typing import TypeVar


@dataclass
class StringGeneric:
    Str = TypeVar("Str", str, str)

    def get_generictype(self) -> Str:
        return self.Str


@dataclass
class IntegerGeneric:
    Int = TypeVar("Int", int, int)

    def get_generictype(self) -> Int:
        return self.Int


@dataclass
class FloatGeneric:
    Float = TypeVar("Float", float, float)

    def get_generictype(self) -> Float:
        return self.Float


@dataclass
class TextGeneric:
    Text = TypeVar("Text", str, str)

    def get_generictype(self) -> Text:
        return self.Text



import time
from typing import ItemsView

from koia.builders.query_builder import QueryBuilder
from koia.builders.utils import check_for_exist_key, sort_fetched_data
from koia.database.connector import Database


class FetchBuilder(QueryBuilder):
    def __init__(self, database: Database):
        super().__init__(database=database)

    def __order_of_commands(self) -> list:
        """Order of attributes"""
        return [
            "innerJoin",
            "where",
            "orWhere",
            "whereIn",
            "whereNotNull",
            "orderBy",
            "take",
        ]

    def __reload_query_conf(self) -> None:
        """Remove query attributes"""
        self.query_conf = {}

    def _formalize_sql(self) -> str:
        """Formalize sql query by set attributes from query_conf"""
        sql_query: str = ""
        order_of_commands: list = self.__order_of_commands()
        query_operations: list = []
        query_items: str
        element: str
        key: int
        elem: str

        for query_items in order_of_commands:
            if query_items == "innerJoin":
                query_operations += [
                    element for element in self.query_conf[query_items]
                ]

            else:
                if "where" in query_items:  # Handle where case
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += check_for_exist_key(
                            query_conf=self.query_conf,
                            query_operations=query_operations,
                            query_items=query_items,
                            exist_key="WHERE",
                            replace_key=",",
                        )

                if "order" in query_items:
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += check_for_exist_key(
                            query_conf=self.query_conf,
                            query_operations=query_operations,
                            query_items=query_items,
                            exist_key="ORDER BY",
                            replace_key=",",
                        )

                if "take" in query_items:
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += check_for_exist_key(
                            query_conf=self.query_conf,
                            query_operations=query_operations,
                            query_items=query_items,
                            exist_key="take",
                            replace_key="",
                        )
                    if len(self.query_conf[query_items]) > 1:
                        raise Exception("You can only once set LIMIT attribute")

        query_operation: str
        for query_operation in query_operations:
            sql_query += f" {query_operation}"

        return sql_query

    def _fetch(self, sql_query) -> dict:
        """Execute sql query

        Output: List[Any] or None
        Return None if sql query is INSERT OR UPDATE"""
        while True:
            try:
                self.database.connection.ping()
                self.database.cursor.execute(sql_query)
                self.database.connection.commit()
                fetch = self.database.cursor.fetchall()
                field_names = [i[0] for i in self.database.cursor.description]
                sorted_data = sort_fetched_data(fetch=fetch, field_names=field_names)

                return {"count": len(sorted_data), "fetch": sorted_data}
            except Exception as e:
                if str(e) == "2013: Lost connection to MySQL server during query":
                    self.__await_db(error=str(e))
                    self.database.reconnect()
                else:
                    raise Exception(e)

    def __await_db(self, error: str):
        """Sleep for a 3 second"""
        print(error)
        print("Try to reconnect after 3 sec. Zzzz...")
        time.sleep(3)

    def get(self) -> dict:
        """Imitize sql SELECT feature"""
        sql_query: str = f"SELECT * FROM {self.query_conf['table']} {self._formalize_sql()}"
        data: dict = self._fetch(sql_query=sql_query)
        self.__reload_query_conf()

        return data

    def insert(self, values: dict) -> None:
        """Imitize sql INSERT feature"""
        values_items: ItemsView = values.items()
        sql_columns: str = ",".join(f"`{column[0]}`" for column in values_items)
        sql_values: str = ",".join(str(value[1]) for value in values_items)

        sql_query: str = f"INSERT INTO {self.query_conf['table']} ({sql_columns}) VALUES ({sql_values})"
        self._fetch(sql_query=sql_query)
        self.__reload_query_conf()

    def update(self, values: dict) -> None:
        """Imitize sql UPDATE feature"""
        sql_query: str = self._formalize_sql()
        update_sql_query: str = f"UPDATE {self.query_conf['table']} SET"

        for key, value_item in enumerate(values.items()):
            if key == len(values) - 1:
                update_sql_query += (
                    f""" {value_item[0]}='{value_item[1].replace("'", " ")}'"""
                )
            else:
                update_sql_query += (
                    f""" {value_item[0]}='{value_item[1].replace("'", " ")}',"""
                )
        sql_query = update_sql_query + sql_query

        self._fetch(sql_query=sql_query)
        self.__reload_query_conf()



import time
from typing import Dict, List, Optional

from koia.builders.utils import sort_fetched_data
from koia.database.connector import Database


class BaseModel:
    __methods__ = ["get"]
    __database__: Database
    __tablename__ = None

    def __await_db(self, error: str) -> None:
        """Sleep for a 3 second"""
        print(error)
        print("Try to reconnect after 3 sec. Zzzz...")
        time.sleep(3)

    def __orderBy(self, orderBy: dict) -> Optional[str]:
        if orderBy:
            sql_pipeline: str = ""
            for order_items in orderBy.items():
                if "ORDER BY" not in sql_pipeline:
                    sql_pipeline += f" ORDER BY {order_items[0]} {order_items[1]} "
                else:
                    sql_pipeline += f" , {order_items[0]} {order_items[1]} "

            return sql_pipeline

    def __filter(self, filter: list = "") -> Optional[str]:
        if len(filter) > 0:
            sql_pipline: str = ""

            for option in filter:
                if "WHERE" not in sql_pipline:
                    sql_pipline += f" WHERE {option} "
                else:
                    sql_pipline += f" AND {option} "

            return sql_pipline

    def __sum_query(self, *args):
        if None not in args:
            return " ".join(args)
        return ""

    def get(
        self, filter: list = "", order_by: dict = None
    ) -> Optional[Dict[str, list]]:
        sql_query: str = ""
        if self.__tablename__ is None:
            raise Exception("Unknown table name")

        columns = [
            column
            for column in dir(self)
            if "__" not in column and column not in self.__methods__
        ]

        try:
            self.__database__.connection.ping()
            sql_query += (
                f"SELECT {', '.join(f'`{column}`' for column in columns)} "
                f"FROM {self.__tablename__} "
            )
            sql_query += self.__sum_query(
                self.__filter(filter=filter), self.__orderBy(orderBy=order_by)
            )
            self.__database__.cursor.execute(sql_query)
            self.__database__.connection.commit()
            fetch: list = self.__database__.cursor.fetchall()
            field_names: list = [i[0] for i in self.__database__.cursor.description]
            sorted_data: List[Dict[str, dict]] = sort_fetched_data(
                fetch=fetch, field_names=field_names
            )

            return {"count": len(sorted_data), "fetch": sorted_data}
        except Exception as e:
            if str(e) == "2013: Lost connection to MySQL server during query":
                self.__await_db(error=str(e))
                self.__database__.reconnect()
            else:
                raise Exception(e)



from typing import Dict, List, Union


def check_for_exist_key(
    query_conf: dict,
    query_operations: list,
    query_items: str,
    exist_key: str,
    replace_key: Union[str, int],
):
    if exist_key == "WHERE":
        return [
            element.replace(exist_key, replace_key)
            if len([elem for elem in query_operations if "WHERE" in elem]) > 0
            or key > 0
            else element
            for key, element in enumerate(query_conf[query_items])
        ]
    elif exist_key == "ORDER BY":
        return [
            element.replace(exist_key, replace_key)
            if len([elem for elem in query_operations if "ORDER BY" in elem]) > 0
            or key > 0
            else element
            for key, element in enumerate(query_conf[query_items])
        ]
    elif exist_key == "take":
        return [element for element in query_conf[query_items]]


def sort_fetched_data(fetch: list, field_names: list) -> List[Dict[str, dict]]:
    sorted_data: List[Dict[str, dict]] = [
        {
            field_names[counter]: fetch[index][counter]
            for counter in range(0, len(field_names))
        }
        for index in range(len(fetch))
    ]
    return sorted_data



from typing import Any, Dict, Generic

# Fix isort warnings later
from koia.types.generic_types import FloatGeneric  # noqa: F841
from koia.types.generic_types import IntegerGeneric  # noqa: F841
from koia.types.generic_types import StringGeneric  # noqa: F841
from koia.types.generic_types import TextGeneric  # noqa: F841


class Column:
    def __init__(
        self,
        column_class: Any,
        pk: bool = False,
        unique: bool = False,
        autoincrement: bool = False,
    ) -> None:
        self.column_class = column_class
        self.pk = pk
        self.unique = unique
        self.autoincrement = autoincrement

    def _track_columns(self) -> None:
        pass

    def get_info(self) -> dict:
        return self.__dict__

    def __set_name__(self, _, value) -> None:
        self.__dict__["column_name"] = value


class String(Generic[StringGeneric.Str]):
    type_of_str: str = "String"

    def __init__(self, max_length: int = 225) -> None:
        self.max_length = max_length

    @staticmethod
    def type_of() -> Dict[str, Any]:
        return {
            "type": type(StringGeneric().get_generictype()),
            "return_value": StringGeneric().get_generictype(),
        }


class Integer(Generic[IntegerGeneric.Int]):
    type_of_str: str = "Integer"

    def __init__(self, max_length: int = 11) -> None:
        self.max_length = max_length

    @staticmethod
    def type_of() -> Dict[str, Any]:
        return {
            "type": type(IntegerGeneric().get_generictype()),
            "return_value": IntegerGeneric().get_generictype(),
        }


class Float(Generic[FloatGeneric.Float]):
    type_of_str: str = "Float"

    def __init__(self, max_length: int = 11) -> None:
        self.max_length = max_length

    @staticmethod
    def type_of() -> Dict[str, Any]:
        return {
            "type": type(FloatGeneric().get_generictype()),
            "return_value": FloatGeneric().get_generictype(),
        }


class Text(Generic[TextGeneric.Text]):
    type_of_str: str = "Text"

    def __init__(self, max_length: int = 100) -> None:
        self.max_length = max_length

    @staticmethod
    def type_of() -> Dict[str, Any]:
        return {
            "type": type(TextGeneric().get_generictype()),
            "return_value": TextGeneric().get_generictype(),
        }



from __future__ import annotations

from typing import Any

from koia.database.connector import Database


class QueryBuilder:
    def __init__(self, database: Database) -> None:
        self.database = database
        self.query_conf: dict = {
            "where": [],
            "whereIn": [],
            "whereNotNull": [],
            "orWhere": [],
            "orderBy": [],
            "take": [],
            "innerJoin": [],
        }

    def __reload_database(self) -> None:
        """Reload database connetion"""
        self.database.reconnect()

    def table(self, table: str) -> QueryBuilder:
        """Set query table"""
        self.query_conf["table"] = table
        return self

    def where(self, column: str, operation: str, value: Any) -> QueryBuilder:
        """Set query WHERE feature"""
        self.query_conf["where"].append(f"WHERE `{column}` {operation} {value}")
        return self

    def whereIn(self, column: str, values: list) -> QueryBuilder:
        """Set query WHERE IN feature"""
        values_string = ",".join(str(value) for value in values)

        self.query_conf["whereIn"].append(f"WHERE `{column}` IN ({values_string})")
        return self

    def whereNotNull(self, column: str) -> QueryBuilder:
        """SET query WHERE NOT NULL feature"""
        self.query_conf["whereNotNull"].append(f"WHERE NOT NULL `{column}`")
        return self

    def orWhere(self, column: str, operation: str, value: Any) -> QueryBuilder:
        """SET query WHERE OR feature"""
        self.query_conf["orWhere"].append(f"OR WHERE `{column}` {operation} {value}")
        return self

    def orderBy(self, column: str, option: str) -> QueryBuilder:
        """Set query ORDER BY feature"""
        self.query_conf["orderBy"].append(f"ORDER BY `{column}` {option}")
        return self

    def take(self, count: int) -> QueryBuilder:
        """Set query LIMIT feature"""
        self.query_conf["take"].append(f"LIMIT {count}")
        return self

    def innerJoin(self, table: str, specification: dict) -> QueryBuilder:
        """Set query INNER JOIN feature"""
        query_string = f"INNER JOIN `{table}` ON "
        counter = 0

        for spec in specification.items():
            if counter == 0:
                query_string += f"({spec[0]} = {spec[1]})"
            else:
                query_string += f" and ({spec[0]} = {spec[1]})"
            counter += 1

        self.query_conf["innerJoin"].append(query_string)
        return self



from dataclasses import dataclass
from typing import Dict, Generic, Tuple, TypeVar

import mysql.connector


@dataclass
class ConnectorGenericTypes:
    T = TypeVar("T", Dict[str, str], Dict[str, int])
    X = TypeVar("X", bool, bool)

    def generic_types(self) -> Tuple[T, X]:
        return self.T, self.X


class Database(Generic[ConnectorGenericTypes.T, ConnectorGenericTypes.X]):
    """
    Create MySQLdbConnectoion instance with cursor
    """

    config: ConnectorGenericTypes.T
    autocommit: ConnectorGenericTypes.X

    def __init__(
        self,
        config: ConnectorGenericTypes.T,
        autocommit: ConnectorGenericTypes.X = False,
    ) -> None:
        """Initialize connection and cursor"""
        self.config = config
        self.automcommit = autocommit
        self.connection = self._connect(autocommit=self.automcommit)
        self.cursor = self._cursor()

    def _connect(self, autocommit: ConnectorGenericTypes.X):
        """Connect to db"""
        _connection = mysql.connector.connect(**self.config)
        _connection.autocommit = autocommit
        return _connection

    def _cursor(self):
        """Create cursor"""
        _cursor = self.connection.cursor(buffered=True)
        return _cursor

    def reconnect(self):
        """Re initialize instance"""
        self.connection.close()
        self.cursor.close()
        self.__init__(config=self.config, autocommit=self.automcommit)

    def close(self):
        """Close instance session"""
        self.connection.close()
        self.cursor.close()
